import cv2

from flip.transformers.element import Element
from flip.transformers.transformer import Transformer


class CreateBoundingBoxes(Transformer):
    def map(self, element: Element) -> Element:
        assert element, "element cannot be None"

        element.tags = self.create(element)

        return element

    def create(self, element):
        array = []
        for obj in element.objects:
            new_x, new_y, new_w, new_h = self.bounding_box(obj.image)
            data = {"name": obj.name, "pos": {}}
            data["pos"]["x"] = (obj.x or 0) + new_x
            data["pos"]["y"] = (obj.y or 0) + new_y
            data["pos"]["w"] = new_w
            data["pos"]["h"] = new_h

            array.append(data)

        return array

    def bounding_box(self, image):
        """
        Args:
          image: image to process its width and height
        Description:
          This function
        Returns:
        """

        # Canny edge detection - edge gradient
        img_gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        gray_filtered = cv2.bilateralFilter(img_gray, 7, 50, 50)
        edged = cv2.Canny(gray_filtered, 60, 120)
        edges_filtered = cv2.Canny(edged, 60, 120)

        # Morphological Transformations
        # applying closing function
        kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (7, 7))
        closed = cv2.morphologyEx(edges_filtered, cv2.MORPH_CLOSE, kernel)

        # Finding_contours
        (cnts, _) = cv2.findContours(
            closed.copy(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
        )

        x1, x2 = img_gray.shape[1], 0
        y1, y2 = img_gray.shape[0], 0
        for c in cnts:

            x, y, w, h = cv2.boundingRect(c)

            if x < x1:
                x1 = x
            if y < y1:
                y1 = y
            if x+w > x2:
                x2 = x+w
            if y+h > y2:
                y2 = y+h

        # new_img = image[y1: y2, x1: x2]

        return x1, y1, x2, y2



import cv2
import numpy as np

from flip.transformers.element import Element
from flip.transformers.transformer import Transformer

from flip import parser


@parser.parseable
class Flip(Transformer):
    """ Flip image of Element

        Parameters
        ----------
        mode : {'random', 'x', 'y'}, default='random'
    """
    _SUPPORTED_MODES = {'random', 'x', 'y'}

    def __init__(self, mode='random'):
        self.mode = mode

        if self.mode not in self._SUPPORTED_MODES:
            raise ValueError("Mode '{0:s}' not supported. ".format(self.mode))

    def map(self, element: Element, parent=None) -> Element:
        assert element, "Element cannot be None"

        if self.mode == 'x':
            direction = 0
        elif self.mode == 'y':
            direction = 1
        else:
            direction = np.random.randint(low=0, high=2)

        if np.random.randint(low=0, high=2) == 0:  # Flip if 0
            element.image = cv2.flip(element.image, direction)

        return element



"""
flip
"""
from . import transformers, utils
import os

APP_ROOT = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

about = {}
with open(f'{APP_ROOT}/flip/__about__.py') as fp:
    exec(fp.read(), about)

__version__ = about['__version__']
__title__ = about['__title__']
__package_name__ = about['__package_name__']
__description__ = about['__description__']
__email__ = about['__email__']
__author__ = about['__author__']
__github__ = about['__github__']
__pypi__ = about['__pypi__']
__license__ = about['__license__']


import os
import json

from flip.transformers.element import Element
from flip.transformers.transformer import Transformer


class CreateJson(Transformer):
    def __init__(self, out_dir: str, name: str = None):
        self.out_dir = out_dir
        self.name = name

    def map(self, element: Element) -> Element:
        assert element, "element cannot be None"

        json_path = os.path.join(self.out_dir, f"{self.name}.json")
        json_data = json.dumps(element.tags)

        with open(json_path, mode="w") as f:
            f.write(json_data)

        return element




import os
import cv2
import numpy as np

from flip.transformers.transformer import Transformer
from flip.transformers.element import Element
from flip import utils


class SaveImage(Transformer):
    def __init__(self, out_dir: str, name: str = None):
        self.out_dir = out_dir
        self.name = name

    def map(self, element: Element) -> Element:
        assert element, "element cannot be None"

        if self.name is None:
            self.name = "f%d" % np.random.randint(0, 5000)

        element.name = self.name
        image_path = os.path.join(
            self.out_dir, f"{self.name}.jpg"
        )  # "{self.name}.{extension}"

        image = utils.inv_channels(element.created_image)
        cv2.imwrite(image_path, image)

        return element



import numpy as np

from flip.transformers.element import Element
from flip.transformers.transformer import Transformer
from flip.utils import rotate_bound, crop_from_contour

class Rotate(Transformer):
    """ Flip image of Element

        Parameters
        ----------
        mode : {'random', '90', 'upside_down'}, default='random'
    """

    _SUPPORTED_MODES = {'random', '90', 'upside_down'}

    def __init__(self, mode='random', min=0, max=360):
        self.mode = mode

        if self.mode not in self._SUPPORTED_MODES:
            raise ValueError("Mode '{0:s}' not supported. ".format(self.mode))

        if self.mode == 'upside_down':
            self.angles = [0, 180]
        elif self.mode == '90':
            self.angles = [0, 90, 180, 270]
        else:
            self.angles = [min, max]

    def map(self, element: Element, parent=None) -> Element:
        assert element, "Element cannot be None"

        if self.mode == 'upside_down' or self.mode == 'ninety':
            angle = np.random.choice(self.angles)
        else:
            angle = np.random.uniform(low=self.angles[0], high=self.angles[1],)

        element.image = rotate_bound(element.image, angle)
        element.image = crop_from_contour(element.image)

        return element



import cv2

# load image PNG format
img = cv2.imread('test2.png', cv2.IMREAD_COLOR)

# canny edge detection - edge gradient
edged = cv2.Canny(img, 10, 250)

# Morphological Transformations
# applying closing function 
kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (7, 7))
closed = cv2.morphologyEx(edged, cv2.MORPH_CLOSE, kernel)

# resources
# Morphological Transformations
# https://opencv-python-tutroals.readthedocs.io/en/latest/py_tutorials/py_imgproc/py_morphological_ops/py_morphological_ops.html

#finding_contours 
(cnts, _) = cv2.findContours(
    closed.copy(),
    cv2.RETR_EXTERNAL,
    cv2.CHAIN_APPROX_SIMPLE)

for c in cnts:
	peri = cv2.arcLength(c, True)
	approx = cv2.approxPolyDP(c, 0.02 * peri, True)
	cv2.drawContours(img, [approx], -1, (0, 255, 0), 2)

idx=0

for c in cnts:
    x,y,w,h = cv2.boundingRect(c)
    # print(x,y)
    if w>50 and h>50:
        idx+=1
        new_img=img[y:y+h,x:x+w]
        cv2.imwrite(str(idx) + '.png', new_img)

        # Window name in which image is displayed
        window_name = 'Object'
        cv2.namedWindow(window_name, cv2.WINDOW_NORMAL)
        # Using cv2.imshow() method
        # Displaying the image  
        
        # cv2.circle(img, (x,y), 2, 
        #     (0,255,0), thickness=6, lineType=8, shift=0)
        copyimg=img.copy()
        ww=new_img.shape[0]
        hh=new_img.shape[1]

        # # print(hh)
        # cv2.rectangle(copyimg,start_point=(x,y),end_point=(x+ww,y+hh),color=(0,255,0),thickness=6)

        cv2.imshow(window_name, img)
        k = cv2.waitKey(0) & 0xFF

        # wait for ESC key to exit 
        if k == 27:
            cv2.destroyAllWindows() 




import os
import json

from flip.transformers.element import Element
from flip.transformers.transformer import Transformer


class CreatePascalVoc(Transformer):
    def __init__(self, out_dir: str, name: str = None):
        self.out_dir = out_dir
        self.name = name

    def map(self, element: Element) -> Element:
        assert element, "element cannot be None"

        xml_path = os.path.join(self.out_dir, f"{self.name}.xml")

        x = self.out_dir.split('/')
        folder = x[len(x)-1]

        f = open(xml_path, "w")
        f.write("")
        f.close()

        f = open(xml_path, "a")
        f.write("<annotation>\n")
        f.write("	<folder>{}</folder>\n".format(folder))
        f.write("	<filename>{}.jpg</filename>\n".format(self.name))
        f.write("	<path>{}/{}.jpg</path>\n".format(self.out_dir, self.name))
        f.write("	<source>\n")
        f.write("		<database>Unknown</database>\n")
        f.write("	</source>\n")
        f.write("	<size>\n")
        f.write("		<width>{}</width>\n".format(element.created_image.shape[1]))
        f.write("		<height>{}</height>\n".format(element.created_image.shape[0]))
        f.write("		<depth>{}</depth>\n".format(element.created_image.shape[2]))
        f.write("	</size>\n")
        f.write("	<segmented>0</segmented>\n")

        for tag in element.tags:
            x_min = int(tag['pos']['x'])
            y_min = int(tag['pos']['y'])
            x_max = x_min + int(tag['pos']['w'])
            y_max = y_min + int(tag['pos']['h'])

            f.write("	<object>\n")
            f.write("		<name>{}</name>\n".format(tag['name']))
            f.write("		<pose>Unspecified</pose>\n")
            f.write("		<truncated>0</truncated>\n")
            f.write("		<difficult>0</difficult>\n")
            f.write("		<bndbox>\n")
            f.write("			<xmin>{}</xmin>\n".format(x_min))
            f.write("			<ymin>{}</ymin>\n".format(y_min))
            f.write("			<xmax>{}</xmax>\n".format(x_max))
            f.write("			<ymax>{}</ymax>\n".format(y_max))
            f.write("		</bndbox>\n")
            f.write("	</object>\n")
        f.write("</annotation>\n")
        f.close()

        return element




__title__ = 'flip-data'
__package_name__ = 'flip'
__version__ = '0.1.9'
__description__ = 'Generate thousands of new 2D images from a small batch of objects and backgrounds.'
__email__ = 'd@linkedai.co'
__author__ = 'Diego Parra'
__github__ = 'https://github.com/LinkedAi/flip'
__pypi__ = 'https://pypi.org/project/flip-data/'
__license__ = 'GPL-3.0'



import typing as tp

from flip import parser
from flip.transformers.element import Element


class Transformer:
    @classmethod
    def parse(cls, **kwargs):
        return cls(**kwargs)

    def __call__(self, elements=None) -> tp.Iterable[Element]:
        assert hasattr(self, "map") != hasattr(self, "process")

        if elements is None:
            elements = ()
        elif isinstance(elements, Element):
            elements = (elements,)

        if hasattr(self, "map"):
            for element in elements:

                elements = self.map(element)

                if isinstance(elements, Element):
                    yield elements
                else:
                    yield from elements

        else:
            yield from self.process(elements)

    def __or__(self, other):
        return Compose([self, other])


@parser.parseable
class Compose(Transformer):
    @classmethod
    def parse(cls, transformers):

        transformers = [parser.parse(transformer) for transformer in transformers]

        return cls(transformers)

    def __init__(self, transformers: tp.List[Transformer]):

        final_transformers = []

        for transformer in transformers:
            if isinstance(transformer, Compose):
                final_transformers += transformer.transformers
            else:
                final_transformers.append(transformer)

        self.transformers = final_transformers

    def process(self, elements: tp.Iterable[Element]) -> tp.Iterable[Element]:

        for transformer in self.transformers:
            elements = transformer(elements)

        return elements


@parser.parseable
class ApplyToObjects(Transformer):
    def __init__(self, transforms: tp.List[Transformer]):
        if hasattr(transforms, "__iter__"):
            self.transforms = transforms
        else:
            self.transforms = [transforms]

    def map(self, element: Element) -> Element:
        for obj in element.objects:
            for system in self.transforms:
                obj = system.map(obj, parent=element)
        return element



import os

from flip.transformers.element import Element
from flip.transformers.transformer import Transformer


class CreateCSV(Transformer):
    def __init__(self, out_dir: str, name: str = None):
        self.out_dir = out_dir
        self.name = name

    def map(self, element: Element) -> Element:
        assert element, "element cannot be None"

        csv_path = os.path.join(self.out_dir, f"{self.name}.csv")
        csv_data = ""

        for tag in element.tags:
            csv_data += f"{tag['name']},{tag['pos']['x']},{tag['pos']['y']},{tag['pos']['w']},{tag['pos']['h']}\n"

        with open(csv_path, mode="w") as f:
            f.write(csv_data)

        return element




import numpy as np

from flip.transformers.element import Element
from flip.transformers.transformer import Transformer


# Only for element with objects
class ObjectsRandomPosition(Transformer):
    """ Set a random position to the objects of Element

        Parameters
        ----------
        mode : {'random', 'percentage'}, default='random'
        overlap : percentage 0 to 1, default=0
                  1 for allowing a complete overlap
                  0 for avoiding overlap
    """
    _SUPPORTED_MODES = {'random', 'percentage'}

    def __init__(
        self,
        mode='random',
        x_min=None,
        x_max=None,
        y_min=None,
        y_max=None,
        overlap=0
    ):
        self.x_min = x_min
        self.x_max = x_max
        self.y_min = y_min
        self.y_max = y_max
        self.overlap = overlap
        self.mode = mode

        if self.mode not in self._SUPPORTED_MODES:
            raise ValueError("Mode '{0:s}' not supported. ".format(self.mode))

    def makemask(self, masks, obj_h, obj_w, low_x, high_x, low_y, high_y,
                 el_h, el_w, factor):
        """ Set a mask

            Parameters
            ----------
            masks : list of masks. Every mask is a dictionary that stores
                    previous objects positions.
            obj_h : high of the new object to insert
            obj_w : width of the new object to insert
            low_x : min x allowed
            high_x : max x allowed
            low_y : min y allowed
            high_y : max y allowed
            el_h: background high
            el_w: background width
            factor: overlapping factor
                    percentage 0 - 1

            Return
            ------
            grid : completed mask
        """
        # create a base mask based on allowed limits

        grid = np.zeros((el_h, el_w))
        grid[low_y:high_y, low_x:high_x] = 1

        # iterate masks - construct and add new masks
        for m in masks:
            c_x = m['w'] / 2
            c_y = m['h'] / 2

            y1 = int(m['y'] + (1 - factor) * c_y)
            x1 = int(m['x'] + (1 - factor) * c_x)
            y2 = int(m['y'] + m['h'] - (1 - factor) * c_y)
            x2 = int(m['x'] + m['w'] - (1 - factor) * c_x)

            y_a = max(0, y1 - obj_h)
            x_a = max(0, x1 - obj_w)
            grid[y_a:y2, x_a:x2] = 0

        return grid

    def map(self, element: Element) -> Element:
        """ locate objects """
        assert element, "Element cannot be None"

        el_h: int = element.image.shape[0]
        el_w: int = element.image.shape[1]

        if self.mode == 'percentage':
            x_min = (
                (self.x_min if self.x_min <= 1 else (self.x_min % 100) / 100) * el_w
                if self.x_min is not None
                else 0
            )
            x_max = (
                (self.x_max if self.x_max <= 1 else (self.x_max % 100) / 100) * el_w
                if self.x_max is not None
                else el_w
            )
            y_min = (
                (self.y_min if self.y_min <= 1 else (self.y_min % 100) / 100) * el_h
                if self.y_min is not None
                else 0
            )
            y_max = (
                (self.y_max if self.y_max <= 1 else (self.y_max % 100) / 100) * el_h
                if self.y_max is not None
                else el_h
            )
        else:
            x_min = self.x_min if self.x_min is not None else 0
            x_max = self.x_max if self.x_max is not None else el_w
            y_min = self.y_min if self.y_min is not None else 0
            y_max = self.y_max if self.y_max is not None else el_h

        masks = []
        for obj in element.objects:
            # object dimension

            obj_h: int = int(obj.image.shape[0])
            obj_w: int = int(obj.image.shape[1])

            # max and min limits for the object with respect to the background
            low_x = int(max(0, x_min))
            high_x = int(min(x_max, el_w - obj_w))
            low_y = int(max(0, y_min))
            high_y = int(min(y_max, el_h - obj_h))

            # overlapping percentage
            factor = self.overlap

            # index_av stores available positions to place the object
            index_av = []

            # find positions to place the object
            # if there is not space the overlapping is increased
            while len(index_av) == 0 and factor < 1:
                factor += 0.1
                params = [masks, obj_h, obj_w, low_x,
                          high_x, low_y, high_y, el_h, el_w, factor]
                grid_available = self.makemask(*params)
                index_av = np.argwhere(grid_available)

            # if despite increasing overlapping there is not space
            # the limits are moved based on a new factor
            factor = 0
            while len(index_av) == 0:
                if high_y == (el_h - obj_h) and high_x == (el_w - obj_w):
                    # when the object doesn't have an available space
                    # despite changing overlapping and limits
                    # the complete background is available
                    index_av = np.argwhere(np.ones((el_h, el_w)))
                low_x = int(max(0, x_min - factor))
                high_x = int(min(x_max + factor, (el_w - obj_w)))
                low_y = int(max(0, y_min - factor))
                high_y = int(min(y_max + factor, el_h - obj_h))
                grid_available = np.zeros((el_h, el_w))
                grid_available[low_y:high_y, low_x:high_x] = 1
                index_av = np.argwhere(grid_available)
                factor += 1

            # random index
            index = np.random.choice(index_av.shape[0])
            obj.y, obj.x = int(index_av[index][0]), int(index_av[index][1])
            # save mask for the inserted object
            masks.append({'x': obj.x, 'y': obj.y, 'w': obj_w, 'h': obj_h})

        return element



import cv2
import numpy as np


def inv_channels(image):
    image[..., :3] = image[..., (2, 1, 0)]
    return image


def rotate_bound(image, angle):
    (h, w) = image.shape[:2]
    (cX, cY) = (w // 2, h // 2)

    scale = 1.0

    # grab the rotation matrix (applying the negative of the
    # angle to rotate clockwise), then grab the sine and cosine
    # (i.e., the rotation components of the matrix)
    M = cv2.getRotationMatrix2D((cX, cY), angle, scale)
    cos = np.abs(M[0, 0])
    sin = np.abs(M[0, 1])

    # compute the new bounding dimensions of the image
    nW = int((h * sin) + (w * cos))
    nH = int((h * cos) + (w * sin))

    # adjust the rotation matrix to take into account translation
    M[0, 2] += (nW / 2) - cX
    M[1, 2] += (nH / 2) - cY

    # distance matrix
    # dst_mat = np.zeros((h, w, 4), np.uint8)

    # perform the actual rotation and return the image
    return cv2.warpAffine(
        image,
        M,
        (nW, nH),
        # dst_mat,
        # flags=cv2.INTER_LINEAR,
        # borderMode=cv2.BORDER_TRANSPARENT,
    )


def overlay_transparent(
    background: np.ndarray,
    overlay: np.ndarray,
    mask: np.ndarray,
    x: int,
    y: int,
    hist_match=False,
    min_object_brightness=0,
):

    background_width = background.shape[1]
    background_height = background.shape[0]

    mask = cv2.GaussianBlur(mask, (7, 7), 0)

    x = x or 0
    y = y or 0

    if x >= background_width or y >= background_height:
        return background

    h, w = overlay.shape[0], overlay.shape[1]

    if x + w > background_width:
        w = background_width - x
        overlay = overlay[:, :w]
        mask = mask[:, :w]

    if y + h > background_height:
        h = background_height - y
        overlay = overlay[:h]
        mask = mask[:h]

    mask = mask / 255.0

    if mask.ndim < background.ndim:
        mask = np.expand_dims(mask, axis=-1)

    if overlay.ndim < background.ndim:
        overlay = np.expand_dims(overlay, axis=-1)

    if background.ndim > 2 and background.shape[2] > 3:
        background = background[..., :3]

    if hist_match:
        overlay = histogram_matching_h(
            overlay, background[y : y + h, x : x + w], min_v=min_object_brightness
        )
    background[y : y + h, x : x + w] = (1.0 - mask) * background[
        y : y + h, x : x + w
    ] + mask * overlay
    return background


def crop_from_contour(image):
    _, _, _, a = cv2.split(image)
    ret, thresh = cv2.threshold(a, 127, 255, 0)
    contours, hierarchy = cv2.findContours(thresh, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)

    index = 0
    for i in range(len(contours)):
        if len(contours[i]) > len(contours[index]):
            index = i
    rect = cv2.boundingRect(contours[index])
    cropped_img = image[rect[1]:(rect[1] + rect[3]), rect[0]:(rect[0] + rect[2])]

    # plt.imshow(cropped_img)
    # plt.show()

    return cropped_img



import cv2

from flip.transformers.element import Element
from flip.transformers.transformer import Transformer
from flip import utils


class Draw(Transformer):
    def map(self, element: Element) -> Element:
        image = element.image.copy()

        img = cv2.GaussianBlur(element.objects[0].image, (7, 7), 0)
        b, g, r, a = cv2.split(img)

        cimg = element.objects[0].image.copy()
        cimg[:, :, 3] = a

        for obj in element.objects:
            image = utils.overlay_transparent(
                background=image,
                overlay=obj.image[..., :-1],
                mask=obj.image[..., -1],
                x=obj.x,
                y=obj.y,
            )

        print("Image created")

        element.created_image = image

        return element



import cv2
import numpy as np

from flip.transformers.element import Element
from flip.transformers.transformer import Transformer


class RandomResize(Transformer):
    """ Random Resize image of Element

        Parameters
        ----------
        mode: {'asymmetric', 'symmetric_w', 'symmetric_h'}, default='asymmetric'
        relation: {'none', 'parent'}
        w_min: when mode='parent' w_min represent a percentage value related to the parent object
        w_max:
        h_min:
        h_max:

        if relation='parent'

        TODO:   asymmetric use w_min, w_max, h_min, h_max / pw_min, pw_max, ph_min, ph_max
                symmetric_w use w_min, w_max / pw_min, pw_max
                symmetric_h use h_min, h_max / ph_min, ph_max
    """
    _SUPPORTED_MODES = {'asymmetric', 'symmetric_w', 'symmetric_h'}

    def __init__(
        self,
        mode='asymmetric',
        w_min=None,
        w_max=None,
        h_min=None,
        h_max=None,
        relation='none',
        w_percentage_min=None,
        w_percentage_max=None,
        h_percentage_min=None,
        h_percentage_max=None,
    ):
        self.w_min = w_min
        self.w_max = w_max
        self.h_min = h_min
        self.h_max = h_max
        self.mode = mode
        self.relation = relation
        self.w_percentage_min = w_percentage_min
        self.w_percentage_max = w_percentage_max
        self.h_percentage_min = h_percentage_min
        self.h_percentage_max = h_percentage_max

        if self.mode not in self._SUPPORTED_MODES:
            raise ValueError("Mode '{0:s}' not supported. ".format(self.mode))

        if self.mode not in self._SUPPORTED_MODES:
            raise ValueError("Relation '{0:s}' not supported. ".format(self.relation))

    def map(self, element: Element, parent=None) -> Element:
        assert element, "Element cannot be None"

        if self.relation == 'parent':
            p_shape = parent.image.shape

            w_min = p_shape[1] * self.w_percentage_min if self.w_percentage_min is not None else 0
            h_min = p_shape[0] * self.h_percentage_min if self.h_percentage_min is not None else 0
            w_max = p_shape[1] * self.w_percentage_max if self.w_percentage_max is not None else p_shape[1]
            h_max = p_shape[0] * self.h_percentage_max if self.h_percentage_max is not None else p_shape[0]
        else:
            w_min = self.w_min if self.w_min is not None else element.image.shape[1]
            h_min = self.h_min if self.h_min is not None else element.image.shape[0]
            w_max = self.w_max if self.w_max is not None else element.image.shape[1]
            h_max = self.h_max if self.h_max is not None else element.image.shape[0]

        if self.mode == 'symmetric_h':
            h = (
                h_min
                if h_min == h_max is not None
                else np.random.randint(low=h_min, high=h_max,)
            )
            w = element.image.shape[1] * (h / element.image.shape[0])
        elif self.mode == 'symmetric_w':
            w = (
                w_min
                if w_min == w_max is not None
                else np.random.randint(low=w_min, high=w_max,)
            )
            h = element.image.shape[0] * (w / element.image.shape[1])
        else:
            w = (
                w_min
                if w_min == w_max is not None
                else np.random.randint(low=w_min, high=w_max,)
            )

            h = (
                h_min
                if h_min == h_max is not None
                else np.random.randint(low=h_min, high=h_max,)
            )

        h = int(h)
        w = int(w)

        element.image = cv2.resize(element.image, (w, h))

        return element



import yaml
import jinja2
import os

TRANSFORMERS = dict()


def parseable(cls):

    name = cls.__name__

    if name in TRANSFORMERS:
        raise ValueError(f"Transformer '{name}' is already registered.")

    TRANSFORMERS[name] = cls

    return cls


def parse(structure):

    if isinstance(structure, list):
        return TRANSFORMERS["Compose"].parse(transformers=structure)
    elif isinstance(structure, dict):
        if len(structure) == 0:
            raise ValueError("Received empty dictionary")
        elif len(structure) > 1:
            raise ValueError(f"Received multiple items: {list(structure)}")

        [(name, kwargs)] = structure.items()

        if kwargs is None:
            kwargs = {}

        return TRANSFORMERS[name].parse(**kwargs)


def loads(content, key="transformers", values=None):

    if values is None:
        values = {}

    env_vars = os.environ.copy()
    env_vars.update(values)

    template = jinja2.Template(content)
    content = template.render(**env_vars)

    structure = yaml.safe_load(content)[key]

    return parse(structure)


def load(buffer, key="transformers", values=None):

    content = buffer.read()

    return loads(content, key=key, values=values)




import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

with open('requirements.txt') as req_file:
    requirements = req_file.read().splitlines()

about = {}
with open('flip/__about__.py') as fp:
    exec(fp.read(), about)

setuptools.setup(
      name=about['__title__'],
      version=about['__version__'],
      license=about['__license__'],
      author=about['__author__'],
      author_email=about['__email__'],
      description=about['__description__'],
      long_description=long_description,
      long_description_content_type='text/markdown',
      url=about['__github__'],
      packages=setuptools.find_packages(),
      classifiers=[
            "Programming Language :: Python :: 3",
            "Operating System :: OS Independent",
      ],
      install_requires=requirements,
      python_requires='>=3.7'
)



class Element:
    """The Element class represents the base class than can be use by a Transform
        Commonly used arguments are:
            - image: Numpy Array with the image of the element
            - name: A name for the element can be use for label it
            - objects: Child Elements of this element, mostly used when the parent Element is a background

        Commonly extra parameters:
            - x, y: To set the position of the element over a parent element
            - tags: To save the information of the generated tags for the child objects
    """

    def __init__(self, **kwargs):

        for key, value in kwargs.items():
            setattr(self, key, value)

    def __getattr__(self, key):
        return None

    def __repr__(self):

        fields = ", ".join(f"{key}={value}" for key, value in vars(self).items())

        return f"Element({fields})"

